@@ -30,6 +30,7 @@
 #include "gkd-secret-unlock.h"
 #include "gkd-secret-util.h"
 
+#include "egg/egg-error.h"
 #include "egg/egg-secure-memory.h"
 
 #include "login/gkd-login.h"
@@ -63,7 +64,7 @@ attributes_for_collection (GP11Object *collection)
 
 	attrs = gp11_object_get (collection, &error, CKA_LABEL, CKA_ID, GP11_INVALID);
 	if (attrs == NULL) {
-		g_warning ("couldn't get attributes for collection: %s", error->message);
+		g_warning ("couldn't get attributes for collection: %s", egg_error_message (error));
 		g_clear_error (&error);
 		return NULL;
 	}
@@ -199,6 +200,11 @@ prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll, gboolean first)
 
 	g_free (label);
 
+	if (gkd_login_is_usable ())
+		gkd_prompt_show_widget (prompt, "auto_unlock_check");
+	else
+		gkd_prompt_hide_widget (prompt, "auto_unlock_check");
+
 	/* Setup the unlock options */
 	if (first) {
 		template = gp11_object_get_template (coll, CKA_G_CREDENTIAL_TEMPLATE, &error);
@@ -206,7 +212,8 @@ prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll, gboolean first)
 			gkd_prompt_set_unlock_options (prompt, template);
 			gp11_attributes_unref (template);
 		} else {
-			g_warning ("couldn't get credential template for collection: %s", error->message);
+			g_warning ("couldn't get credential template for collection: %s",
+			           egg_error_message (error));
 			g_clear_error (&error);
 		}
 	}
@@ -228,9 +235,9 @@ check_locked_collection (GP11Object *collection, gboolean *locked)
 
 	value = gp11_object_get_data (collection, CKA_G_LOCKED, &n_value, &error);
 	if (value == NULL) {
-		if (error->code != CKR_OBJECT_HANDLE_INVALID)
+		if (!g_error_matches (error, GP11_ERROR, CKR_OBJECT_HANDLE_INVALID))
 			g_warning ("couldn't check locked status of collection: %s",
-			           error->message);
+			           egg_error_message (error));
 		return FALSE;
 	}
 
@@ -240,41 +247,35 @@ check_locked_collection (GP11Object *collection, gboolean *locked)
 }
 
 static void
-attach_credential_to_login (GP11Object *collection, GP11Object *cred)
+attach_unlock_to_login (GP11Object *collection, GkdSecretSecret *master)
 {
-	GError *error = NULL;
+	DBusError derr = DBUS_ERROR_INIT;
 	GP11Attributes *attrs;
-	gpointer value;
-	gsize n_value;
+	GP11Object *cred;
 	gchar *location;
 	gchar *label;
 
 	g_assert (GP11_IS_OBJECT (collection));
-	g_assert (GP11_IS_OBJECT (cred));
 
+	/* Relevant information for the unlock item */
 	attrs = attributes_for_collection (collection);
 	g_return_if_fail (attrs);
-
 	location = location_string_for_attributes (attrs);
 	label = label_string_for_attributes (attrs);
 	gp11_attributes_unref (attrs);
 
-	value = gp11_object_get_data_full (cred, CKA_VALUE, egg_secure_realloc, NULL, &n_value, &error);
-	if (value) {
-		if (g_utf8_validate (value, n_value, NULL))
-			gkd_login_attach_secret (label, value, "keyring", location, NULL);
-		else
-			g_warning ("couldn't save non utf-8 unlock credentials in login keyring");
-		egg_secure_clear (value, n_value);
-		egg_secure_free (value);
-
-	} else {
-		g_warning ("couldn't read unlock credentials to save in login keyring: %s", error->message);
-		g_clear_error (&error);
-	}
-
+	attrs = gkd_login_attach_make_attributes (label, "keyring", location, NULL);
 	g_free (location);
 	g_free (label);
+
+	cred = gkd_secret_session_create_credential (master->session, NULL, attrs, master, &derr);
+	gp11_attributes_unref (attrs);
+	g_object_unref (cred);
+
+	if (!cred) {
+		g_warning ("couldn't save unlock password in login collection: %s", derr.message);
+		dbus_error_free (&derr);
+	}
 }
 
 static void
@@ -292,8 +293,10 @@ authenticate_collection (GkdSecretUnlock *self, GP11Object *collection, gboolean
 	DBusError derr = DBUS_ERROR_INIT;
 	GkdSecretSecret *master;
 	GP11Attributes *template;
+	GP11Attribute *attr;
 	GP11Object *cred;
 	gboolean transient;
+	gboolean result;
 
 	g_assert (GKD_SECRET_IS_UNLOCK (self));
 	g_assert (GP11_IS_OBJECT (collection));
@@ -318,39 +321,45 @@ authenticate_collection (GkdSecretUnlock *self, GP11Object *collection, gboolean
 	gkd_prompt_get_unlock_options (GKD_PROMPT (self), template);
 
 	/* If it's supposed to save non-transient, then we override that */
-	if (!gp11_attributes_find_boolean (template, CKA_GNOME_TRANSIENT, &transient))
-		transient = TRUE;
+	attr = gp11_attributes_find (template, CKA_GNOME_TRANSIENT);
+	if (attr != NULL) {
+		transient = gp11_attribute_get_boolean (attr);
+		gp11_attribute_clear (attr);
+		gp11_attribute_init_boolean (attr, CKA_GNOME_TRANSIENT, TRUE);
+	}
 
 	cred = gkd_secret_session_create_credential (master->session, NULL, template, master, &derr);
-	gkd_secret_secret_free (master);
+	g_object_unref (cred);
 
 	if (cred) {
 		/* Save it to the login keyring */
 		if (!transient)
-			attach_credential_to_login (collection, cred);
-		g_object_unref (cred);
+			attach_unlock_to_login (collection, master);
 
 		/* Save away the unlock options for next time */
 		gp11_object_set_template (collection, CKA_G_CREDENTIAL_TEMPLATE, template, NULL);
 		gp11_attributes_unref (template);
 
 		*locked = FALSE;
-		return TRUE; /* Operation succeeded, and unlocked */
+		result = TRUE; /* Operation succeeded, and unlocked */
 
 	} else {
 		gp11_attributes_unref (template);
 		if (dbus_error_has_name (&derr, INTERNAL_ERROR_DENIED)) {
 			dbus_error_free (&derr);
 			*locked = TRUE;
-			return TRUE; /* Operation succeded, although not unlocked*/
+			result = TRUE; /* Operation succeded, although not unlocked*/
 
 		} else {
 			g_warning ("couldn't create credential for collection: %s",
 			           derr.message);
 			dbus_error_free (&derr);
-			return FALSE; /* Operation failed */
+			result = FALSE; /* Operation failed */
 		}
 	}
+
+	gkd_secret_secret_free (master);
+	return result;
 }
 
 /* -----------------------------------------------------------------------------
@@ -634,10 +643,10 @@ gkd_secret_unlock_with_password (GP11Object *collection, const guchar *password,
 
 	cred = gp11_session_create_object_full (session, attrs, NULL, &error);
 	if (cred == NULL) {
-		if (error->code == CKR_PIN_INCORRECT) {
+		if (g_error_matches (error, GP11_ERROR, CKR_PIN_INCORRECT)) {
 			dbus_set_error_const (derr, INTERNAL_ERROR_DENIED, "The password was incorrect.");
 		} else {
-			g_message ("couldn't create credential: %s", error->message);
+			g_message ("couldn't create credential: %s", egg_error_message (error));
 			dbus_set_error_const (derr, DBUS_ERROR_FAILED, "Couldn't use credentials");
 		}
 		g_clear_error (&error);
