@@ -30,6 +30,7 @@
 #include "pkcs11/pkcs11g.h"
 #include "pkcs11/pkcs11i.h"
 
+#include "egg/egg-error.h"
 #include "egg/egg-unix-credentials.h"
 
 #include <sys/types.h>
@@ -82,7 +83,7 @@ typedef struct _CallState {
 	GckRpcMessage *req;
 	GckRpcMessage *resp;
 	void *allocated;
-	CK_ULONG appid;
+	CK_G_APPLICATION application;
 } CallState;
 
 static int
@@ -97,7 +98,10 @@ call_init (CallState *cs)
 		gck_rpc_message_free (cs->resp);
 		return 0;
 	}
-	
+
+	memset (&cs->application, 0, sizeof (cs->application));
+	cs->application.applicationData = cs;
+
 	cs->allocated = NULL;
 	return 1;
 }
@@ -780,9 +784,8 @@ rpc_C_Finalize (CallState *cs)
 {
 	CK_SLOT_ID_PTR slots;
 	CK_ULONG n_slots, i;
-	CK_SLOT_ID appartment;
 	CK_RV ret;
-	
+
 	debug (("C_Finalize: enter"));
 
 	assert (cs);
@@ -792,26 +795,26 @@ rpc_C_Finalize (CallState *cs)
 	 * We don't actually C_Finalize lower layers, since this would finalize
 	 * for all appartments, client applications. Anyway this is done by 
 	 * the code that loaded us.
-	 * 
+	 *
 	 * But we do need to cleanup resources used by this client, so instead
-	 * we call C_CloseAllSessions for each appartment for this client.
+	 * we call C_CloseAllSessions for each slot for this client application.
 	 */
-	
-	ret = (pkcs11_module->C_GetSlotList) (TRUE, NULL, &n_slots);
-	if (ret == CKR_OK) {
-		slots = calloc (n_slots, sizeof (CK_SLOT_ID));
-		if (slots == NULL) {
-			ret = CKR_DEVICE_MEMORY;
-		} else {
-			ret = (pkcs11_module->C_GetSlotList) (TRUE, slots, &n_slots);
-			for (i = 0; ret == CKR_OK && i < n_slots; ++i) {
-				appartment = CK_GNOME_MAKE_APPARTMENT (slots[i], cs->appid);
-				ret = (pkcs11_module->C_CloseAllSessions) (appartment);
+
+	if (cs->application.applicationId) {
+		ret = (pkcs11_module->C_GetSlotList) (TRUE, NULL, &n_slots);
+		if (ret == CKR_OK) {
+			slots = calloc (n_slots, sizeof (CK_SLOT_ID));
+			if (slots == NULL) {
+				ret = CKR_DEVICE_MEMORY;
+			} else {
+				ret = (pkcs11_module->C_GetSlotList) (TRUE, slots, &n_slots);
+				for (i = 0; ret == CKR_OK && i < n_slots; ++i)
+					ret = (pkcs11_module->C_CloseAllSessions) (slots[i] | cs->application.applicationId);
+				free (slots);
 			}
-			free (slots);
 		}
 	}
-	
+
 	debug (("ret: %d", ret));
 	return ret;
 }
@@ -852,7 +855,6 @@ rpc_C_GetSlotInfo (CallState *cs)
 
 	BEGIN_CALL (C_GetSlotInfo);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 	PROCESS_CALL ((slot_id, &info));
 		OUT_SLOT_INFO (info);
 	END_CALL;
@@ -868,7 +870,6 @@ rpc_C_GetTokenInfo (CallState *cs)
 	
 	BEGIN_CALL (C_GetTokenInfo);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 	PROCESS_CALL ((slot_id, &info));
 		OUT_TOKEN_INFO (info);
 	END_CALL;
@@ -885,7 +886,6 @@ rpc_C_GetMechanismList (CallState *cs)
 
 	BEGIN_CALL (C_GetMechanismList);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 		IN_ULONG_BUFFER (mechanism_list, count);
 	PROCESS_CALL ((slot_id, mechanism_list, &count));
 		OUT_ULONG_ARRAY (mechanism_list, count);
@@ -903,7 +903,6 @@ rpc_C_GetMechanismInfo (CallState *cs)
 	
 	BEGIN_CALL (C_GetMechanismInfo);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 		IN_ULONG (type);
 	PROCESS_CALL ((slot_id, type, &info));
 		OUT_MECHANISM_INFO (info);
@@ -922,7 +921,6 @@ rpc_C_InitToken (CallState *cs)
 
 	BEGIN_CALL (C_InitToken);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 		IN_BYTE_ARRAY (pin, pin_len);
 		IN_STRING (label);
 	PROCESS_CALL ((slot_id, pin, pin_len, label));
@@ -940,7 +938,6 @@ rpc_C_WaitForSlotEvent (CallState *cs)
 	BEGIN_CALL (C_WaitForSlotEvent);
 		IN_ULONG (flags);
 	PROCESS_CALL ((flags, &slot_id, NULL));
-		slot_id = CK_GNOME_APPARTMENT_SLOT (slot_id);
 		OUT_ULONG (slot_id);
 	END_CALL;
 }
@@ -956,9 +953,9 @@ rpc_C_OpenSession (CallState *cs)
 
 	BEGIN_CALL (C_OpenSession);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
 		IN_ULONG (flags);
-	PROCESS_CALL ((slot_id, flags, NULL, NULL, &session));
+		flags |= CKF_G_APPLICATION_SESSION;
+	PROCESS_CALL ((slot_id, flags, &cs->application, NULL, &session));
 		OUT_ULONG (session);
 	END_CALL;
 }
@@ -984,7 +981,7 @@ rpc_C_CloseAllSessions (CallState *cs)
 
 	BEGIN_CALL (C_CloseAllSessions);
 		IN_ULONG (slot_id);
-		slot_id = CK_GNOME_MAKE_APPARTMENT (slot_id, cs->appid);
+		slot_id |= cs->application.applicationId;
 	PROCESS_CALL ((slot_id));
 	END_CALL;
 }
@@ -1022,7 +1019,6 @@ rpc_C_GetSessionInfo (CallState *cs)
 	BEGIN_CALL (C_GetSessionInfo);
 		IN_ULONG (session);
 	PROCESS_CALL ((session, &info));
-		info.slotID  = CK_GNOME_APPARTMENT_SLOT (info.slotID);
 		OUT_SESSION_INFO (info);
 	END_CALL;
 }
@@ -2089,10 +2085,7 @@ run_dispatch_loop (int sock)
 		gck_rpc_warn ("couldn't read socket credentials");
 		return;
 	}
-	
-	/* The client application */
-	cs.appid = pid;
-	
+
 	/* Setup our buffers */
 	if (!call_init (&cs)) {
 		gck_rpc_warn ("out of memory");
@@ -2141,7 +2134,18 @@ run_dispatch_loop (int sock)
 		   !write_all (sock, cs.resp->buffer.buf, cs.resp->buffer.len))
 			break;
 	}
-	
+
+	/*
+	 * Close all sessions for the client application
+	 *
+	 * EXTENSION: In our extended application PKCS#11 model
+	 * C_CloseAllSessions accepts an application identifier as well
+	 * as slot ids. Calling with an application identifier closes all
+	 * sessions for just that application identifier.
+	 */
+	if (cs.application.applicationId)
+		(pkcs11_module->C_CloseAllSessions) (cs.application.applicationId);
+
 	call_uninit (&cs);
 }
 
@@ -2175,7 +2179,7 @@ typedef struct _DispatchState {
 static int pkcs11_socket = -1;
 
 /* The unix socket path, that we listen on */
-static char pkcs11_socket_path[MAXPATHLEN] = { 0, };
+static char *pkcs11_socket_path = NULL;
 
 /* A linked list of dispatcher threads */
 static DispatchState *pkcs11_dispatchers = NULL;
@@ -2220,7 +2224,7 @@ gck_rpc_layer_accept (void)
 	
 	ds->thread = g_thread_create (run_dispatch_thread, &(ds->socket), TRUE, &error);
 	if (!ds->thread) {
-		gck_rpc_warn ("couldn't start thread: %s", error && error->message ? error->message : "");
+		gck_rpc_warn ("couldn't start thread: %s", egg_error_message (error));
 		close (new_fd);
 		free (ds);
 		return;
@@ -2231,7 +2235,55 @@ gck_rpc_layer_accept (void)
 }
 
 int
-gck_rpc_layer_initialize (const char *prefix, CK_FUNCTION_LIST_PTR module)
+gck_rpc_layer_initialize (CK_FUNCTION_LIST_PTR module)
+{
+	assert (module);
+
+	/* cannot be called more than once */
+	assert (!pkcs11_module);
+	pkcs11_module = module;
+	return 1;
+}
+
+void
+gck_rpc_layer_uninitialize (void)
+{
+	DispatchState *ds, *next;
+
+	if (!pkcs11_module)
+		return;
+
+	/* Close our main listening socket */
+	if (pkcs11_socket != -1)
+		close (pkcs11_socket);
+	pkcs11_socket = -1;
+
+	/* Delete our unix socket */
+	if(pkcs11_socket_path) {
+		unlink (pkcs11_socket_path);
+		free (pkcs11_socket_path);
+		pkcs11_socket_path = NULL;
+	}
+
+	/* Stop all of the dispatch threads */
+	for (ds = pkcs11_dispatchers; ds; ds = next) {
+		next = ds->next;
+
+		/* Forcibly shutdown the connection */
+		if (ds->socket)
+			shutdown (ds->socket, SHUT_RDWR);
+		g_thread_join (ds->thread);
+
+		/* This is always closed by dispatch thread */
+		assert (ds->socket == -1);
+		free (ds);
+	}
+
+	pkcs11_module = NULL;
+}
+
+int
+gck_rpc_layer_startup (const char *prefix)
 {
 	struct sockaddr_un addr;
 	int sock;
@@ -2240,23 +2292,26 @@ gck_rpc_layer_initialize (const char *prefix, CK_FUNCTION_LIST_PTR module)
 	GCK_RPC_CHECK_CALLS ();
 #endif
 
-	assert (module);
 	assert (prefix);
 
 	/* cannot be called more than once */
-	assert (!pkcs11_module);
 	assert (pkcs11_socket == -1);
 	assert (pkcs11_dispatchers == NULL);
-	
-	snprintf (pkcs11_socket_path, sizeof (pkcs11_socket_path), 
-	          "%s/socket.pkcs11", prefix);
+
+	free (pkcs11_socket_path);
+	pkcs11_socket_path = malloc (strlen (prefix) + strlen ("/pkcs11") + 1);
+	if (pkcs11_socket_path == NULL) {
+		gck_rpc_warn ("couldn't allocate memory");
+		return -1;
+	}
+	sprintf (pkcs11_socket_path, "%s/pkcs11", prefix);
 
 	sock = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (sock < 0) {
 		gck_rpc_warn ("couldn't create pkcs11 socket: %s", strerror (errno));
 		return -1;
 	}
-	
+
 	memset(&addr, 0, sizeof(addr));
 	addr.sun_family = AF_UNIX;
 	unlink (pkcs11_socket_path);
@@ -2266,27 +2321,23 @@ gck_rpc_layer_initialize (const char *prefix, CK_FUNCTION_LIST_PTR module)
 		                  pkcs11_socket_path, strerror (errno));
 		return -1;
 	}
-	
+
 	if (listen (sock, 128) < 0) {
 		gck_rpc_warn ("couldn't listen on pkcs11 socket: %s: %s", 
 		                  pkcs11_socket_path, strerror (errno));
 		return -1;
 	}
-	
-	pkcs11_module = module;
+
 	pkcs11_socket = sock;
 	pkcs11_dispatchers = NULL;
-	
+
 	return sock;
 }
 
 void
-gck_rpc_layer_uninitialize (void)
+gck_rpc_layer_shutdown (void)
 {
 	DispatchState *ds, *next;
-	
-	if (!pkcs11_module)
-		return;
 
 	/* Close our main listening socket */
 	if (pkcs11_socket != -1) 
@@ -2294,9 +2345,11 @@ gck_rpc_layer_uninitialize (void)
 	pkcs11_socket = -1;
 
 	/* Delete our unix socket */
-	if(pkcs11_socket_path[0])
+	if(pkcs11_socket_path) {
 		unlink (pkcs11_socket_path);
-	pkcs11_socket_path[0] = 0;
+		free (pkcs11_socket_path);
+		pkcs11_socket_path = NULL;
+	}
 
 	/* Stop all of the dispatch threads */
 	for (ds = pkcs11_dispatchers; ds; ds = next) {
@@ -2311,6 +2364,4 @@ gck_rpc_layer_uninitialize (void)
 		assert (ds->socket == -1);
 		free (ds);
 	}
-	
-	pkcs11_module = NULL;
 }
